@@ -173,7 +173,6 @@ static gboolean
 model_node_has_children (GtkTreeModel *tree_model, GNode *node)
 {
 	ModelNode *mnode;
-	AccessibleTreeModel *model = ACCESSIBLE_TREE_MODEL (tree_model);
 	Accessible *accessible;
 	gboolean retval;
 
@@ -203,8 +202,8 @@ model_node_sync_children (GtkTreeModel *tree_model, GNode *node)
 	GSList *children, *l;
 	ModelNode *mnode;
 	AccessibleTreeModel *model = ACCESSIBLE_TREE_MODEL (tree_model);
-	AccessibleTable *accessible_table;
-	gboolean use_table = FALSE;
+	AccessibleTable *accessible_table = NULL;
+	long n_rows, n_cols;
 
 	g_return_if_fail (node != NULL);
 
@@ -213,22 +212,25 @@ model_node_sync_children (GtkTreeModel *tree_model, GNode *node)
 	if (mnode->state & NODE_GOT_CHILDREN)
 		return;
 
-	if (use_table_if) {
+	if (use_table_if)
 		accessible_table = Accessible_getTable (mnode->accessible);
 
-		if (accessible_table)
-			use_table = TRUE;
-	} 			
-
  retry_on_reenter:
 	mnode->state |= NODE_GOT_CHILDREN | NODE_FETCHING_CHILDREN;
 
 	model->stamp++;
 	children = NULL;
 
-	if (use_table) {
-		for (i = 0; max_children ? i < max_children : TRUE; i++) {
-			for (j = 0; max_children ? j < max_children : TRUE; j++) {
+	if (accessible_table) {
+		if (!max_children) {
+			n_rows = AccessibleTable_getNRows (accessible_table);
+			n_cols = AccessibleTable_getNColumns (accessible_table);
+		} else {
+			n_rows = max_children;
+			n_cols = max_children;
+		}
+		for (i = 0; i < n_rows; i++) {
+			for (j = 0; j < n_cols; j++) {
 				Accessible *accessible;
 
 				accessible = AccessibleTable_getAccessibleAt (accessible_table, i, j);
@@ -691,7 +693,8 @@ report_global_event (const AccessibleEvent *event, void *user_data)
 
 		role_txt = Accessible_getRoleName (active_descendant);
 
-		g_print ("New active descendant: %p Role: %s Index: %d\n", active_descendant, role_txt, event->detail1);
+		g_print ("New active descendant: %p Role: %s Index: %ld\n",
+			 active_descendant, role_txt, event->detail1);
 		SPI_freeString (role_txt);
 	}
 }
