@@ -34,6 +34,8 @@
 #include "accessible-listener.h"
 #include "accessible-tree-model.h"
 
+#define BLINK_INTERVAL_MS 200
+
 extern void poke (Accessible *accessible);
 extern int max_children;
 extern gboolean use_table_if;
@@ -94,6 +96,14 @@ typedef struct {
 	gboolean      poker_died;
 } Poker;
 
+typedef struct {
+	gint count;
+	gint x;
+	gint y;
+	gint w;
+	gint h;
+} BlinkClosure;
+
 enum {
 	POKER_SELECTION_ICON,
 	POKER_SELECTION_NAME,
@@ -154,6 +164,12 @@ update_relation_set (Poker *poker, AccessibleRelation **relations)
 		{ SPI_RELATION_MEMBER_OF,      "MEMBER_OF" },
 		{ SPI_RELATION_NODE_CHILD_OF,  "NODE_CHILD_OF" },
 		{ SPI_RELATION_EXTENDED,       "EXTENDED" },
+		{ SPI_RELATION_FLOWS_TO,       "FLOWS_TO" },
+		{ SPI_RELATION_FLOWS_FROM,     "FLOWS_FROM" },
+		{ SPI_RELATION_SUBWINDOW_OF,   "SUBWINDOW_OF" },
+		{ SPI_RELATION_EMBEDS,         "EMBEDS" },
+		{ SPI_RELATION_EMBEDDED_BY,    "EMBEDDED_BY" },
+		{ SPI_RELATION_POPUP_FOR,      "POPUP_FOR" },
 		{ 0, NULL }
 	};
 
@@ -312,6 +328,62 @@ component_grab_focus_clicked (GtkButton *button, Poker *poker)
 	AccessibleComponent_unref (component);
 }
 
+static gboolean
+blink_draw_rectangle (gpointer user_data)
+{
+	static GtkWidget *blink_widget = NULL;
+	BlinkClosure *blink_data = user_data;
+
+	GdkDisplay *display = gdk_display_get_default ();
+	GdkScreen *screen = gdk_display_get_default_screen (display);
+	GdkWindow *root_window = gdk_screen_get_root_window (screen);
+ 
+	GdkGC *gc = gdk_gc_new (root_window);
+
+	/* grab mouse for the invisible widget when we are blinking */
+	if (!blink_widget) {
+		blink_widget = gtk_invisible_new_for_screen (screen);
+		gtk_widget_show (blink_widget);
+		gtk_grab_add (blink_widget);
+	}
+
+	/* set subwindow mode, function and line attributes for our GC */
+	gdk_gc_set_subwindow (gc, GDK_INCLUDE_INFERIORS);
+	gdk_gc_set_function (gc, GDK_INVERT);
+	gdk_gc_set_line_attributes (gc, 3, GDK_LINE_SOLID,
+				    GDK_CAP_BUTT, GDK_JOIN_MITER);
+	gdk_draw_rectangle (root_window, gc, FALSE,
+			    blink_data->x, blink_data->y,
+			    blink_data->w, blink_data->h);
+	g_object_unref (gc);
+
+	--(blink_data->count);
+
+	/* return FALSE will destroy the timer. */
+	if (blink_data->count <= 0) {
+		gtk_grab_remove (blink_widget);
+		gtk_widget_destroy (blink_widget);
+		blink_widget = NULL;
+		g_free (blink_data);
+		return FALSE;
+	}
+	return TRUE;
+}
+
+static void
+component_blink_extents (long int x, long int y, long int w, long int h)
+{
+	BlinkClosure *blink_data = g_new0 (BlinkClosure, 1);
+
+	blink_data->count = 4;  /* should be an even number */
+	blink_data->x = x;
+	blink_data->y = y;
+	blink_data->w = w;
+	blink_data->h = h;
+
+	g_timeout_add (BLINK_INTERVAL_MS, blink_draw_rectangle, blink_data);
+}
+
 static void
 update_if_component (Poker *poker, AccessibleComponent *component)
 {
@@ -335,8 +407,14 @@ update_if_component (Poker *poker, AccessibleComponent *component)
 		{ 0, NULL }
 	};
 
-	AccessibleComponent_getExtents (
-		component, &x, &y, &w, &h, poker->ctype);
+	/* make our blinks */
+	AccessibleComponent_getExtents (component, &x, &y, &w, &h,
+					SPI_COORD_TYPE_SCREEN);
+	component_blink_extents (x, y, w, h);
+
+	if (poker->ctype != SPI_COORD_TYPE_SCREEN)
+		AccessibleComponent_getExtents (component, &x, &y, &w, &h,
+						poker->ctype);
 
 	widget = glade_xml_get_widget (poker->xml, "component_if_position");
 	txt = g_strdup_printf ("%ld, %ld", x, y);
@@ -402,6 +480,33 @@ action_row_activated_cb (GtkTreeView       *tree_view,
 	action_take_clicked (NULL, poker);
 }
 
+static void
+link_activated_cb (GtkTreeView       *tree_view,
+		   GtkTreePath       *path,
+		   GtkTreeViewColumn *column,
+		   Poker             *poker)
+{
+	gint index;
+	AccessibleHypertext *hypertext;
+	AccessibleHyperlink *link;
+	AccessibleAction *action;
+
+	index = gtk_tree_path_get_indices (path) [0];
+
+	g_return_if_fail (poker->selected);
+
+	hypertext = Accessible_getHypertext (poker->selected);
+	g_return_if_fail (hypertext);
+	link = AccessibleHypertext_getLink (hypertext, index);
+	g_return_if_fail (link);
+	action = Accessible_getAction (link);
+	AccessibleAction_doAction (action, 0);
+
+	Accessible_unref (action);
+	Accessible_unref (link);
+	Accessible_unref (hypertext);
+}
+
 static void
 update_if_action (Poker *poker, AccessibleAction *action)
 {
@@ -799,12 +904,12 @@ update_if_selection (Poker *poker, AccessibleSelection *selection)
 	AccessibleTable *table;
 	GtkTreeSelection *view_sel;
 	gboolean use_table = FALSE;
+	long n_rows, n_cols;
 
 	parent = Accessible_queryInterface (
 		selection, "IDL:Accessibility/Accessible:1.0");
 	g_return_if_fail (parent != NULL);
 
-
 	gtk_list_store_clear (poker->selection_store);
 	view_sel = gtk_tree_view_get_selection (poker->selection_view);
 
@@ -819,14 +924,21 @@ update_if_selection (Poker *poker, AccessibleSelection *selection)
 			use_table = TRUE;
 	}
 	if (use_table) {
-		for (j = 0; max_children ? j < max_children : TRUE; i++) {
-			for (k = 0; max_children ? k < max_children : TRUE; i++) {
-			
+                if (!max_children) {
+                        n_rows = AccessibleTable_getNRows (table);
+                        n_cols = AccessibleTable_getNColumns (table);
+                } else {
+                        n_rows = max_children;
+                        n_cols = max_children;
+                }
+                for (j = 0; j < n_rows; j++) {
+                        for (k = 0; k < n_cols; k++) {
+
 				child = AccessibleTable_getAccessibleAt (table, j, k);
 				if (!child)
 					break;
 
-				i++;
+				i = AccessibleTable_getIndexAt (table, j, k);
 				process_accessible (poker, selection, child, view_sel, i);
 			}
 		}
@@ -916,23 +1028,36 @@ update_if_hypertext (Poker *poker, AccessibleHypertext *hypertext)
 	long i;
 	long num_links;
 	GtkTreeIter iter;
+	AccessibleText *text;
 
 	num_links = AccessibleHypertext_getNLinks (hypertext);
 	gtk_list_store_clear (poker->hypertext_store);
 
+	text = Accessible_getText (hypertext);
+	g_return_if_fail (text);
+
 	for (i = 0; i < num_links; i++) {
 		AccessibleHyperlink *link;
 		long n_anchors;
+		long start_index, end_index;
+		char *txt;
 
 		link = AccessibleHypertext_getLink (hypertext, i);
 		n_anchors = AccessibleHyperlink_getNAnchors (link);
-		Accessible_unref (link);
+
+		AccessibleHyperlink_getIndexRange (link, &start_index, &end_index);
+		txt = AccessibleText_getText (text, start_index, end_index);
 		gtk_list_store_append (poker->hypertext_store, &iter);	
 		gtk_list_store_set (poker->hypertext_store, &iter,
 				    0, i,
-                                    1, n_anchors,
-                                    -1);
+				    1, n_anchors,
+				    2, txt,
+				    -1);
+		SPI_freeString (txt);
+
+		Accessible_unref (link);
 	}
+	Accessible_unref (text);
 }
 
 static const struct {
@@ -1417,7 +1542,7 @@ poke (Accessible *accessible)
 
 	{ /* hypertext interface */
 		poker->hypertext_store = gtk_list_store_new (
-			2, G_TYPE_INT, G_TYPE_INT);
+			3, G_TYPE_INT, G_TYPE_INT, G_TYPE_STRING);
 
 		widget = glade_xml_get_widget (
 			poker->xml, "hypertext_if_links");
@@ -1435,6 +1560,14 @@ poke (Accessible *accessible)
 			gtk_cell_renderer_text_new (), "text",
 			1, NULL);
 
+		gtk_tree_view_insert_column_with_attributes (
+			GTK_TREE_VIEW (widget), 2, "Name",
+			gtk_cell_renderer_text_new (), "text",
+			2, NULL);
+
+		g_signal_connect (widget, "row_activated",
+				  G_CALLBACK (link_activated_cb),
+				  poker);
 	}
 
 	{ /* main frame buttons */
